# Copyright 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import binascii
import ctypes
from ctypes import (
    CFUNCTYPE, Structure, c_uint8, c_uint32, c_long, c_void_p,
    c_uint64, c_char_p, c_int, c_uint16, POINTER, Union, byref)
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import (
    WjhDropRawInfo,
    WjhL1DropAggregate,
    WjhDropAggregate)
import random
from scapy.all import (raw, Ether, IP, IPv6, TCP, UDP)
import subprocess
import time

AGENTSTATUS_SUCCESS = 0
AGENTSTATUS_FAILURE = 1

WJH_USER_CHANNEL_TAILDROP_E = 0
WJH_USER_CHANNEL_CYCLIC_E = 1
WJH_USER_CHANNEL_AGGREGATE_E = 2

WJH_DROP_REASON_GROUP_BUFFER_E = 0
WJH_DROP_REASON_GROUP_ACL_E = 1
WJH_DROP_REASON_GROUP_L1_E = 2
WJH_DROP_REASON_GROUP_L2_E = 3
WJH_DROP_REASON_GROUP_ROUTER_E = 4
WJH_DROP_REASON_GROUP_TUNNEL_E = 5

WJH_USER_CHANNEL_MODE_PUSH_E = 0
WJH_USER_CHANNEL_MODE_PULL_E = 1

WJH_SEVERITY_ALL_E = 0
WJH_SEVERITY_NOTICE_E = 1
WJH_SEVERITY_WARNING_E = 2
WJH_SEVERITY_ERROR_E = 3

severity_to_str = {
    WJH_SEVERITY_NOTICE_E: 'Notice',
    WJH_SEVERITY_WARNING_E: 'Warning',
    WJH_SEVERITY_ERROR_E: 'Error'
}

WJH_ACL_BINDING_POINT_INGRESS = 0
WJH_ACL_BINDING_POINT_EGRESS = 1
WJH_ACL_BINDING_POINT_RIF_INGRESS = 2
WJH_ACL_BINDING_POINT_RIF_EGRESS = 3

acl_bp_to_str = {
    WJH_ACL_BINDING_POINT_INGRESS: 'Ingress ACL',
    WJH_ACL_BINDING_POINT_EGRESS: 'Egress ACL',
    WJH_ACL_BINDING_POINT_RIF_INGRESS: 'RIF Ingress ACL',
    WJH_ACL_BINDING_POINT_RIF_EGRESS: 'RIF Egress ACL'
}

WjhDeactiveCb = CFUNCTYPE(None)


class WjhInitParam(Structure):
    _fields_ = [("force", c_uint8), ("deactive_cb", WjhDeactiveCb), ("bandwidth_percent", c_uint8),
                ("log_cb", c_void_p), ("conf_xml_path", c_char_p)]


class WjhDropCounter(Structure):
    _fields_ = [("dropped", c_uint64), ("received", c_uint64)]


class WjhTimespec(Structure):
    _fields_ = [("tv_sec", c_long), ("tv_nsec", c_long)]


class WjhAggregateTimestamp(Structure):
    _fields_ = [("first_timestamp", WjhTimespec), ("last_timestamp", WjhTimespec)]


class WjhDropReason(Structure):
    _fields_ = [("id", c_uint32),
                ("reason", c_char_p),
                ("severity", c_int),
                ("description", c_char_p)]


class WjhDropBasicAttr(Structure):
    _fields_ = [("recirculation_port", c_uint32),
                ("truncate_size", c_uint16)]


class WjhDropAdvancedAttr(Structure):
    _fields_ = [("trap_probability", c_uint32),
                ("trap_id", c_uint32),
                ("ptp_disabled", c_uint32)]


class WjhBufferDropAttr(Structure):
    _fields_ = [("basic_attr", WjhDropBasicAttr),
                ("advanced_attr", WjhDropAdvancedAttr)]


class WjhDropReasonGroupAttrUnion(Union):
    _fields_ = [("buffer_drop", WjhBufferDropAttr)]


class WjhDropReasonGroupAttr(Structure):
    _fields_ = [("attr", WjhDropReasonGroupAttrUnion)]


class WjhL2DropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhL2DropRawCb = CFUNCTYPE(c_int, POINTER(WjhL2DropRawInfo), POINTER(c_uint32))


class WjhRouterDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhRouterDropRawCb = CFUNCTYPE(c_int, POINTER(WjhRouterDropRawInfo), POINTER(c_uint32))


class WjhTunnelDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhTunnelDropRawCb = CFUNCTYPE(c_int, POINTER(WjhTunnelDropRawInfo), POINTER(c_uint32))


class WjhBufferDropRawInfo(Structure):
    _fields_ = [("packet", ctypes.c_void_p),
                ("packet_size", ctypes.c_uint32),
                ("ingress_port", ctypes.c_uint32),
                ("egress_port", ctypes.c_uint32),
                ("tc", ctypes.c_uint8),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", ctypes.c_uint8),
                ("ingress_lag", ctypes.c_uint32)]


WjhBufferDropRawCb = CFUNCTYPE(c_int, POINTER(WjhBufferDropRawInfo), POINTER(c_uint32))


class WjhL1DropAggregateKey(Structure):
    _fields_ = [("ingress_port", c_uint32)]


class WjhL1DropAggregateData(Structure):
    _fields_ = [("is_port_up", c_uint8),
                ("port_down_reason", c_char_p),
                ("description", c_char_p),
                ("state_change_count", c_uint64),
                ("symbol_error_count", c_uint64),
                ("crc_error_count", c_uint64),
                ("timestamp", WjhAggregateTimestamp)]


WjhL1DropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhL1DropAggregateKey), POINTER(WjhL1DropAggregateData), POINTER(c_uint32))

class WjhAggregateFiveTuples(Structure):
    _fields_ = [("sip", c_char_p),
                ("dip", c_char_p),
                ("proto", c_uint8),
                ("sport", c_uint32),
                ("dport", c_uint32)]

class WjhRouterDropAggregateKey(Structure):
    _fields_ = [("five_tuples", WjhAggregateFiveTuples),
                ("reason", c_char_p)]

class WjhRouterDropAggregateData(Structure):
    _fields_ = [("timestamp", WjhAggregateTimestamp),
                ("count", c_uint32)]


WjhRouterDropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhRouterDropAggregateKey), POINTER(WjhRouterDropAggregateData), POINTER(c_uint32))


class WjhAclDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("binding_point", c_int),
                ("rule_id", c_uint64),
                ("acl_name", c_char_p),
                ("rule", c_char_p),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhAclDropRawCb = CFUNCTYPE(c_int, POINTER(WjhAclDropRawInfo), POINTER(c_uint32))


class WjhUserChannelAttr(Structure):
    _fields_ = [("polling_interval", c_uint32),
                ("mode", c_int)]


class RawCallback(Union):
    _fields_ = [("buffer", WjhBufferDropRawCb),
                ("acl", WjhAclDropRawCb),
                ("router", WjhRouterDropRawCb),
                ("L2", WjhL2DropRawCb),
                ("tunnel", WjhTunnelDropRawCb)]


class AggregateCallback(Union):
    _fields_ = [("buffer", c_void_p),
                ("acl", c_void_p),
                ("router", c_void_p),
                ("L2", c_void_p),
                ("L1", WjhL1DropAggregateCb),
                ("tunnel", c_void_p)]


class WjhDropCallbacks(Structure):
    _fields_ = [("drop_reason_group", c_int),
                ("raw_cb", RawCallback),
                ("aggregate_cb", AggregateCallback)]



class WjhClient(object):

    def __init__(self, conf, logger, is_mlx=True):
        self._conf = conf
        self._logger = logger
        self.wjh_lib = None
        self.portmap = {}
        self.agg_cache = {}
        self.sdk_pid = 0
        self.is_mlx = is_mlx


    def create_user_channel(self, channel_type):
        user_channel = c_uint32()
        if self.wjh_lib.wjh_user_channel_create(channel_type,
                                                byref(user_channel)) != 0:
            self._logger.error("Failed to create user channel")
        else:
            self._logger.info("create channel id %d" % (user_channel.value))

        return user_channel.value


    def get_counter(self, user_channel_id):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_counter_get(user_channel_id,
                    byref(drop_counter), 0) != 0:

            self._logger.error("Failed to get counter")
        else:
            self._logger.info("channel[%d]: drop_counter [%d], recieve_counter [%d]" % (
                              user_channel_id, drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def get_global_counter(self):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_global_counter_get(byref(drop_counter), 0) != 0:
            self._logger.error("Failed to get global counter")
        else:
            self._logger.error("global: drop_counter [%d], recieve_counter [%d]" % (
                               drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def configure_drop_reason_group(self, drop_reason_group, channel_id):
        if self.wjh_lib.wjh_drop_reason_group_bind(drop_reason_group, channel_id) != 0:
            self._logger.error("Failed to bind drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_enable(drop_reason_group, 0) != 0:
            self._logger.error("Failed to enable drop reason group")


    def deconfigure_drop_reason_group(self, drop_reason_group):
        if self.wjh_lib.wjh_drop_reason_group_disable(drop_reason_group, 0) != 0:
            self._logger.error("Failed to disable drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_unbind(drop_reason_group) != 0:
            self._logger.error("Failed to unbind drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_deinit(drop_reason_group) != 0:
            self._logger.error("Failed to deinit drop reason group")


    def wjh_callback(self, channel_id, aggregation_channel_id):
        if self.get_port_map() == AGENTSTATUS_FAILURE:
            return

        l2_cb = WjhDropCallbacks()
        l2_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L2_E
        l2_cb.raw_cb.L2 = WjhL2DropRawCb(self.wjh_l2_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L2_E, byref(attr), byref(l2_cb)) != 0:
            self._logger.error("Failed to init L2 drop reason group")

        router_cb = WjhDropCallbacks()
        router_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ROUTER_E
        router_cb.raw_cb.router = WjhRouterDropRawCb(self.wjh_router_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ROUTER_E, byref(attr), byref(router_cb)) != 0:
            self._logger.error("Failed to init router drop reason group")

        tunnel_cb = WjhDropCallbacks()
        tunnel_cb.drop_reason_group = WJH_DROP_REASON_GROUP_TUNNEL_E
        tunnel_cb.raw_cb.tunnel = WjhTunnelDropRawCb(self.wjh_tunnel_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_TUNNEL_E, byref(attr), byref(tunnel_cb)) != 0:
            self._logger.error("Failed to init tunnel drop reason group")

        acl_cb = WjhDropCallbacks()
        acl_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ACL_E
        acl_cb.raw_cb.acl = WjhAclDropRawCb(self.wjh_acl_drop_raw_cb)
        attr = c_uint32()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ACL_E, byref(attr), byref(acl_cb)) != 0:
            self._logger.error("Failed to init acl drop reason group")

        buffer_cb = WjhDropCallbacks()
        buffer_cb.drop_reason_group = WJH_DROP_REASON_GROUP_BUFFER_E
        buffer_cb.raw_cb.buffer = WjhBufferDropRawCb(self.wjh_buffer_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        attr.attr.buffer_drop.basic_attr.recirculation_port = 0x10100
        attr.attr.buffer_drop.advanced_attr.trap_probability = 8
        # TBD: sdk support not ready, it's crashing if enabled
        #rc = wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_BUFFER_E, byref(attr), byref(buffer_cb))
        #assert 0 == rc, "Failed to init buffer drop reason group"
        l1_cb = WjhDropCallbacks()
        l1_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L1_E
        l1_cb.aggregate_cb.L1 = WjhL1DropAggregateCb(self.wjh_l1_drop_aggregate_cb)
        attr = c_uint32()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L1_E, byref(attr), byref(l1_cb)) != 0:
            self._logger.error("Failed to init L1 drop reason group")

        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L2_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ROUTER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_TUNNEL_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ACL_E, channel_id)
        #configure_drop_reason_group(WJH_DROP_REASON_GROUP_BUFFER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L1_E, aggregation_channel_id)

        while True:
            time.sleep(15)
            self.flush_agg_cache()


    def wjh_deactive_cb(self):
        self._logger.info("This WJH client is deactivated by another WJH client")
        return

    def get_port_map(self):
        cmd = '/usr/bin/cat /cumulus/switchd/run/hashtable/1'
        try:
            output = subprocess.check_output(cmd.split())
        except Exception:
            self._logger.debug('Cannot get portmap')
            return AGENTSTATUS_FAILURE

        port = ''
        lid = ''
        for l in output.splitlines():
            if 'port-info' in l:
                port = l.split()[1]
            if 'lid' in l:
                lid = l.split()[1]
                self.portmap[lid] = port

        return AGENTSTATUS_SUCCESS


    def start_dummy_client(self):
        class _raw_info(object):
            def __init__(self, reason='', drop_type='', ingress_port='',
                         egress_port='', traffic_class=0, acl_rule_id=0,
                         acl_bind_point=0, acl_name='', acl_rule='', smac='',
                         dmac='', sip='', dip='', proto=0, sport=0, dport=0,
                         fts=0.0, lts=0.0, agg_count=0):
                self.drop_agg = WjhDropAggregate()
                self.drop_agg.reason = reason
                self.drop_agg.drop_type = drop_type
                self.drop_agg.ingress_port = ingress_port
                self.drop_agg.egress_port = egress_port
                self.drop_agg.traffic_class = traffic_class
                self.drop_agg.acl_rule_id = acl_rule_id
                self.drop_agg.acl_bind_point = acl_bind_point
                self.drop_agg.acl_name = acl_name
                self.drop_agg.acl_rule = acl_rule
                self.drop_agg.smac = smac
                self.drop_agg.dmac = dmac
                self.drop_agg.sip = sip
                self.drop_agg.dip = dip
                self.drop_agg.proto = proto
                self.drop_agg.sport = sport
                self.drop_agg.dport = dport
                self.drop_agg.agg_count = agg_count
                self.drop_agg.first_timestamp = fts
                self.drop_agg.last_timestamp = lts

        self._logger.info('Started DUMMY wjh agent')
        if 'spine' in HOSTNAME or 'tor' in HOSTNAME:
            seed = int(HOSTNAME.split('-')[1])
            random.seed(seed)
            t1 = time.time()
            fts = t1 - (random.random() * 20)
            lts = t1 + (random.random() * 20)

            if 'spine' in HOSTNAME:
                raw_info = _raw_info(reason='Blackhole route', drop_type='Router',
                                     ingress_port='swp1', smac='00:01:02:03:04:05',
                                     dmac='06:07:08:09:0a:0b', proto=17, sport=12345,
                                     dport=45643, sip='192.168.0.1', dip='192.168.0.2',
                                     fts=fts, lts=lts,
                                     agg_count=random.randrange(100, 2000))
            if 'tor' in HOSTNAME:
                raw_info = _raw_info(reason='Source MAC equals destination MAC',
                                     drop_type='L2', ingress_port='swp2',
                                     smac='00:01:02:03:04:05',
                                     dmac='00:01:02:03:04:05', proto=6, sport=55520,
                                     dport=10456, sip='45.1.0.13', dip='3.0.6.2',
                                     fts=fts, lts=lts,
                                     agg_count=random.randrange(200, 1000))
            drop_agg = raw_info.drop_agg
            key = (drop_agg.reason, drop_agg.ingress_port, drop_agg.sip, drop_agg.dip,
                   drop_agg.proto, drop_agg.sport, drop_agg.dport, drop_agg.smac,
                   drop_agg.dmac)
            self.agg_cache[key] = drop_agg

            # add an L1 drop also
            L1drop = WjhL1DropAggregate()
            L1drop.ingress_port = 'swp{}'.format(random.randrange(1, 8))
            L1drop.is_port_up = False
            L1drop.port_down_reason = 'Admin down'
            L1drop.description = ''
            L1drop.state_change_count = random.randrange(1, 10)
            L1drop.symbol_error_count = random.randrange(100, 1000)
            L1drop.crc_error_count = random.randrange(200, 2000)
            L1drop.first_timestamp = fts
            L1drop.last_timestamp = lts
            L1drop.save()

        while True:
            time.sleep(15)
            self.flush_agg_cache()


    def start_wjh_client(self):
        if not self.is_mlx:
            return self.start_dummy_client()

        try:
            ctypes.CDLL('libsxapi.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxadviser.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxaclhelper.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libscew.so', mode=ctypes.RTLD_GLOBAL)
            self.wjh_lib = ctypes.CDLL('libwjh.so', mode=ctypes.RTLD_GLOBAL)
        except OSError:
            self._logger.error("WJH lib not present")
            return AGENTSTATUS_FAILURE

        #wjh_helper = ctypes.CDLL('libwjhhelper.so', mode=ctypes.RTLD_GLOBAL)

        try:
            cmd = '/usr/bin/pidof sx_sdk'
            sdk_pid = subprocess.check_output(cmd.split())
            self.sdk_pid = int(sdk_pid)
        except Exception:
            self._logger.debug("sx_sdk not alive")
            return AGENTSTATUS_FAILURE

        init_param = WjhInitParam()
        init_param.force = 1
        init_param.deactive_cb = WjhDeactiveCb(self.wjh_callback)
        init_param.bandwidth_percent = 0
        init_param.log_cb = 0
        init_param.conf_xml_path = 0

        if self.wjh_lib.wjh_init(byref(init_param)) != 0:
            self._logger.error("wjh_init failed")
            return AGENTSTATUS_FAILURE

        channel_id = self.create_user_channel(WJH_USER_CHANNEL_CYCLIC_E)
        aggregation_channel_id = self.create_user_channel(WJH_USER_CHANNEL_AGGREGATE_E)

        self.wjh_callback(channel_id, aggregation_channel_id)

        if self.wjh_lib.wjh_user_channel_destroy(channel_id) != 0:
            self._logger.error("Failed to destroy user channel")

        if self.wjh_lib.wjh_user_channel_destroy(aggregation_channel_id) != 0:
            self._logger.error("Failed to destroy user channel")

        if self.wjh_lib.wjh_deinit() != 0:
            self._logger.error("Failed to deinit wjh service lib")

        self._logger.info("wjh lib is deinitialized")
        return AGENTSTATUS_SUCCESS


    #
    # Raw callback handlers
    #
    def wjh_raw_drop_fill(self, info):
        drop_reason = WjhDropRawInfo()
        drop_reason.reason_id = info.drop_reason.id
        drop_reason.reason = info.drop_reason.reason
        drop_reason.description = info.drop_reason.description
        drop_reason.severity = severity_to_str[info.drop_reason.severity]
        lid = '0x%x' % info.ingress_port
        drop_reason.ingress_port = self.portmap.get(lid) or lid
        drop_reason.timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)

        packet_size = info.packet_size
        packet = list((packet_size * ctypes.c_ubyte).from_address(info.packet))

        drop_reason.packet_size = packet_size
        if (packet_size > 0):
            drop_reason.packet = ''.join('%02x' % i for i in packet)

        return drop_reason


    def parse_7_tuples(self, packet, packet_size):
        tup = {'smac': '00:00:00:00:00:00',
               'dmac': '00:00:00:00:00:00',
               'src': '0.0.0.0',
               'dst': '0.0.0.0',
               'sport': 0,
               'dport': 0,
               'proto': 0}

        if packet_size == 0:
            return tup

        pkt = list((packet_size * ctypes.c_ubyte).from_address(packet))
        try:
            p = raw(binascii.a2b_hex(''.join('%02x' % i for i in pkt)))
            eth = Ether(p)
        except Exception as e:
            return tup

        tup['smac'] = eth.src
        tup['dmac'] = eth.dst

        l3_found = False
        iph = None

        try:
            iph = eth[IP]
            l3_found = True
            tup['proto'] = iph.proto
        except Exception as e:
            self._logger.debug('{}'.format(e))

        if not l3_found:
            try:
                iph = eth[IPv6]
                l3_found = True
                tup['proto'] = iph.nh
            except Exception as e:
                self._logger.debug('{}'.format(e))

        if l3_found:
            tup['src'] = iph.src
            tup['dst'] = iph.dst

            try:
                udph = eth[UDP]
                tup['sport'] = udph.sport
                tup['dport'] = udph.dport
            except Exception as e:
                self._logger.debug('{}'.format(e))

            try:
                tcph = eth[TCP]
                tup['sport'] = tcph.sport
                tup['dport'] = tcph.dport
            except Exception as e:
                self._logger.debug('{}'.format(e))

        self._logger.debug('smac: {}, dmac: {} -- src: {}, dst: {}, proto {}, sport {}, dport {}'.format(
                           tup['smac'], tup['dmac'], tup['src'], tup['dst'], tup['proto'], tup['sport'], tup['dport']))
        return tup


    def wjh_raw_drop_fill_aggregate(self, drop_type, info, egress_port='',
                                    traffic_class=0, acl_rule_id=0):
        drop_reason = WjhDropAggregate()
        if drop_type == 'ACL':
            drop_reason.reason = '{}'.format(acl_bp_to_str.get(info.binding_point) or '')
            drop_reason.acl_bind_point = info.binding_point
            drop_reason.acl_name = info.acl_name
            drop_reason.acl_rule = info.rule
        else:
            drop_reason.reason = info.drop_reason.reason
        drop_reason.drop_type = drop_type
        p_tuples = self.parse_7_tuples(info.packet, info.packet_size)
        drop_reason.sip = p_tuples['src']
        drop_reason.dip = p_tuples['dst']
        drop_reason.proto = p_tuples['proto']
        drop_reason.sport = p_tuples['sport']
        drop_reason.dport = p_tuples['dport']
        drop_reason.smac = p_tuples['smac']
        drop_reason.dmac = p_tuples['dmac']
        drop_reason.egress_port = egress_port
        drop_reason.traffic_class = traffic_class
        drop_reason.acl_rule_id = acl_rule_id
        lid = '0x%x' % info.ingress_port
        drop_reason.ingress_port = self.portmap.get(lid) or lid

        key = (drop_reason.reason, drop_reason.ingress_port, drop_reason.sip, drop_reason.dip,
               drop_reason.proto, drop_reason.sport, drop_reason.dport, drop_reason.smac,
               drop_reason.dmac)

        agg = self.agg_cache.get(key)
        if agg:
            agg.agg_count += 1
            agg.last_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
        else:
            drop_reason.agg_count = 1
            drop_reason.first_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
            drop_reason.last_timestamp = drop_reason.first_timestamp
            self.agg_cache[key] = drop_reason

        return drop_reason


    def flush_agg_cache(self):
        for key, agg in self.agg_cache.items():
            self._logger.info("Aggregate count: {} first {} last {}".format(agg.agg_count, agg.first_timestamp, agg.last_timestamp))
            agg.save()

        self.agg_cache.clear()


    def wjh_l2_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("l2 drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str[raw_info_list_p[i].drop_reason.severity],
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('L2', raw_info_list_p[i])

        return 0


    def wjh_router_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("router drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str[raw_info_list_p[i].drop_reason.severity],
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Router', raw_info_list_p[i])

        return 0


    def wjh_tunnel_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("tunnel drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str[raw_info_list_p[i].drop_reason.severity],
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Tunnel', raw_info_list_p[i])

        return 0


    def wjh_buffer_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("buffer drop reason group: ingress port [0x%x], size[%d], reason[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Buffer', raw_info_list_p[i],
                                   egress_port=raw_info_list_p[i].egress_port,
                                   traffic_class=raw_info_list_p[i].tc)

        return 0


    def wjh_acl_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            self._logger.debug("acl drop reason group: ingress port [0x%x], size[%d], rule_id[0x%x], binding_point[0x%x], acl_name[%s], rule{%s}, timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                     raw_info_list_p[i].packet_size,
                     raw_info_list_p[i].rule_id,
                     raw_info_list_p[i].binding_point,
                     raw_info_list_p[i].acl_name,
                     raw_info_list_p[i].rule,
                     raw_info_list_p[i].timestamp.tv_sec,
                     raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('ACL', raw_info_list_p[i], acl_rule_id=raw_info_list_p[i].rule_id)
        return 0


    #
    # Aggregate callback handlers
    #
    def wjh_agg_drop_fill(self, key, data):
        drop = WjhDropAggregate()
        drop.reason = key.reason
        drop.sip = key.five_tuples.sip
        drop.dip = key.five_tuples.dip
        drop.proto = key.five_tuples.proto
        drop.sport = key.five_tuples.sport
        drop.dport = key.five_tuples.dport
        drop.agg_count = data.count
        drop.first_timestamp = float(data.timestamp.first_timestamp.tv_sec) + float(
                                     data.timestamp.first_timestamp.tv_nsec / 1000000000.0)
        drop.last_timestamp = float(data.timestamp.last_timestamp.tv_sec) + float(
                                           data.timestamp.last_timestamp.tv_nsec / 1000000000.0)
        return drop


    def wjh_acl_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        # ACL aggregate callback is not very useful as it does not have 5 tuples
        pass


    def wjh_buf_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        # buffer drop aggregate is not useful as it does not have traffic_class
        # info
        pass

    def wjh_l1_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        self._logger.info("")
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            lid = '0x%x' % key_list_p[i].ingress_port
            ingress_port = self.portmap.get(lid) or lid
            '''
            self._logger.info("port %s: is_port_up: %u, port_down_reason: %s, description: %s, state_change_count: %u" %
                  (ingress_port, data_list_p[i].is_port_up, data_list_p[i].port_down_reason,
                   data_list_p[i].description, data_list_p[i].state_change_count))
            self._logger.info("           symbol_error_count: %u, crc_error_count: %u" %
                  (data_list_p[i].symbol_error_count, data_list_p[i].crc_error_count))
            self._logger.info("           timestamp[%ld.%ld, %ld.%ld]" % (data_list_p[i].timestamp.first_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.first_timestamp.tv_nsec, data_list_p[i].timestamp.last_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.last_timestamp.tv_nsec))
            '''
            # L1 drop callback data is quite different from the others
            l1drop = WjhL1DropAggregate()
            l1drop.ingress_port = '0x%x' % key_list_p[i].ingress_port
            l1drop.is_port_up = data_list_p[i].is_port_up
            l1drop.port_down_reason = data_list_p[i].port_down_reason
            l1drop.description = data_list_p[i].description
            l1drop.state_change_count = data_list_p[i].state_change_count
            l1drop.symbol_error_count = data_list_p[i].symbol_error_count
            l1drop.crc_error_count = data_list_p[i].crc_error_count
            l1drop.first_timestamp = float(data_list_p[i].timestamp.first_timestamp.tv_sec) + float(
                                           data_list_p[i].timestamp.first_timestamp.tv_nsec / 1000000000.0)
            l1drop.last_timestamp = float(data_list_p[i].timestamp.last_timestamp.tv_sec) + float(
                                          data_list_p[i].timestamp.last_timestamp.tv_nsec / 1000000000.0)
            olddrops = WjhL1DropAggregate().memmodel()
            olddrop = olddrops.query.get(hostname=HOSTNAME, ingress_port=ingress_port)
            if (olddrop and olddrop.is_port_up == l1drop.is_port_up and
                olddrop.port_down_reason == l1drop.port_down_reason and
                olddrop.state_change_count == l1drop.state_change_count and
                olddrop.symbol_error_count == l1drop.symbol_error_count and
                olddrop.crc_error_count == l1drop.crc_error_count):
                continue

            l1drop.save()

        return 0


    def wjh_rtr_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            '''
            self._logger.info("sip %s, dip %s, proto %u, sport %u, dport %u, reason %s" %
                  (key_list_p[i].five_tuples.sip, key_list_p[i].five_tuples.dip, key_list_p[i].five_tuples.proto, key_list_p[i].five_tuples.sport, key_list_p[i].five_tuples.dport, key_list_p[i].reason))
            self._logger.info("count: %u, timestamp[%ld.%ld, %ld.%ld]" % (data_list_p[i].count,
                                                              data_list_p[i].timestamp.first_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.first_timestamp.tv_nsec, data_list_p[i].timestamp.last_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.last_timestamp.tv_nsec))
            '''
            rtrdrop = self.wjh_agg_drop_fill(key_list_p[i], data_list_p[i])
            rtrdrop.save()
        return 0

    def wjh_l2_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            l2drop = self.wjh_agg_drop_fill(key_list_p[i], data_list_p[i])
            smac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].smac))
            dmac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].dmac))
            l2drop.smac = ''.join('%02x' % i for i in smac)
            l2drop.dmac = ''.join('%02x' % i for i in dmac)
            l2drop.save()

    def wjh_tun_drop_aggregate_cb(self, key_list_p, data_list_p, keys_cnt_p):
        list_size = keys_cnt_p.contents.value
        for i in range(list_size):
            tundrop = self.wjh_agg_drop_fill(key_list_p[i], data_list_p[i])
            smac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].smac))
            dmac = list((6 * ctypes.c_ubyte).from_address(key_list_p[i].dmac))
            tundrop.smac = ''.join('%02x' % i for i in smac)
            tundrop.dmac = ''.join('%02x' % i for i in dmac)
            tundrop.save()

    def _run():
        pass
